@@ -18,11 +18,11 @@
  * along with bcal.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-#include <ctype.h>
-#include <quadmath.h>
 #include <stdio.h>
-#include <string.h>
+#include <ctype.h>
 #include <unistd.h>
+#include <readline/history.h>
+#include <readline/readline.h>
 #include "dslib.h"
 #include "log.h"
 
@@ -36,39 +36,189 @@
 #define FLOAT_BUF_LEN 128
 #define FLOAT_WIDTH 40
 #define ARRAY_SIZE(x) (sizeof(x) / sizeof((x)[0]))
+#define MAX_BITS 128
+#define _ALIGNMENT_MASK 0xF
+
+typedef unsigned char bool;
+typedef unsigned int uint;
+typedef unsigned long ulong;
+typedef unsigned long long ull;
+typedef long double maxfloat_t;
 
 #ifdef __SIZEOF_INT128__
 typedef __uint128_t maxuint_t;
-typedef __float128 maxfloat_t;
 #else
 typedef __uint64_t maxuint_t;
-typedef double maxfloat_t;
 #endif
 
-typedef unsigned char bool;
-typedef unsigned long ulong;
-typedef unsigned long long ull;
-
 typedef struct {
 	ulong c;
 	ulong h;
 	ulong s;
 } t_chs;
 
-static char *VERSION = "1.7";
+static char *VERSION = "1.9";
 static char *units[] = {"b", "kib", "mib", "gib", "tib", "kb", "mb", "gb", "tb"};
 
+static char *FAILED = "1";
+static char *PASSED = "\0";
+static char *curexpr = NULL;
+
 static char uint_buf[UINT_BUF_LEN];
 static char float_buf[FLOAT_BUF_LEN];
-static int unitless;
 
 static int minimal;
-int cur_loglevel = INFO;
+static int repl;
+
+static Data lastres = {"\0", 0};
+
+static int cur_loglevel = INFO;
+static char *logarr[] = {"ERROR", "WARNING", "INFO", "DEBUG"};
+
+static void debug_log(const char *func, int level, const char *format, ...)
+{
+	va_list ap;
+
+	va_start(ap, format);
+
+	if (level < 0 || level > DEBUG)
+		return;
+
+	if (level <= cur_loglevel) {
+		if (cur_loglevel == DEBUG) {
+			fprintf(stderr, "%s(), %s: ", func, logarr[level]);
+			vfprintf(stderr, format, ap);
+		} else {
+			fprintf(stderr, "%s: ", logarr[level]);
+			vfprintf(stderr, format, ap);
+		}
+	}
+
+	va_end(ap);
+}
+
+/*
+ * Just a safe strncpy(3)
+ * Always null ('\0') terminates if both src and dest are valid pointers.
+ * Returns the number of bytes copied including terminating null byte.
+ */
+static size_t bstrlcpy(char *dest, const char *src, size_t n)
+{
+	static ulong *s, *d;
+	static size_t len, blocks;
+	static const uint lsize = sizeof(ulong);
+	static const uint _WSHIFT = (sizeof(ulong) == 8) ? 3 : 2;
+
+	if (!src || !dest || !n)
+		return 0;
+
+	len = strlen(src) + 1;
+	if (n > len)
+		n = len;
+	else if (len > n)
+		/* Save total number of bytes to copy in len */
+		len = n;
+
+	/*
+	 * To enable -O3 ensure src and dest are 16-byte aligned
+	 * More info: http://www.felixcloutier.com/x86/MOVDQA.html
+	 */
+	if ((n >= lsize) && (((ulong)src & _ALIGNMENT_MASK) == 0 && ((ulong)dest & _ALIGNMENT_MASK) == 0)) {
+		s = (ulong *)src;
+		d = (ulong *)dest;
+		blocks = n >> _WSHIFT;
+		n &= lsize - 1;
+
+		while (blocks) {
+			*d = *s;
+			++d, ++s;
+			--blocks;
+		}
+
+		if (!n) {
+			dest = (char *)d;
+			*--dest = '\0';
+			return len;
+		}
+
+		src = (char *)s;
+		dest = (char *)d;
+	}
+
+	while (--n && (*dest = *src))
+		++dest, ++src;
+
+	if (!n)
+		*dest = '\0';
+
+	return len;
+}
+
+/*
+ * Try to evaluate en expression using bc
+ */
+static int try_bc()
+{
+	pid_t pid;
+	int pipe_pc[2], pipe_cp[2], ret;
+
+	if (!curexpr)
+		return -1;
+
+	log(DEBUG, "expression: \"%s\"\n", curexpr);
+
+	if (pipe(pipe_pc) == -1 || pipe(pipe_cp) == -1) {
+		printf("Could not pipe\n");
+		exit(EXIT_FAILURE);
+	}
+
+	pid = fork();
+
+	if (pid == -1) {
+		log(ERROR, "fork() failed!\n");
+		return -1;
+	}
+
+	if (pid == 0) { /* child */
+		close(STDOUT_FILENO);
+		close(STDIN_FILENO);
+		close(pipe_pc[1]); // Close writing end
+		close(pipe_cp[0]); // Close reading end
+
+		dup2(pipe_pc[0], STDIN_FILENO); // Take stdin from parent
+		dup2(pipe_cp[1], STDOUT_FILENO); // Give stdout to parent
+		dup2(pipe_cp[1], STDERR_FILENO); // Give stderr to parent
+
+		ret = execlp("bc", "bc", "-q", (char*) NULL);
+		log(ERROR, "execlp() failed!\n");
+		exit(ret);
+	}
+
+	/* parent */
+	char buffer[128] = "";
+	ret = write(pipe_pc[1], "scale=5\n", 8);
+	ret = write(pipe_pc[1], curexpr, strlen(curexpr));
+	ret = write(pipe_pc[1], "\n", 1);
+	ret = read(pipe_cp[0], buffer, sizeof(buffer));
+
+	if (buffer[0] != '(') {
+		printf("%s", buffer);
+		bstrlcpy(lastres.p, buffer, NUM_LEN);
+		/* remove newline */
+		lastres.p[strlen(lastres.p) - 1] = '\0';
+		lastres.unit = 0;
+		log(DEBUG, "result: %s %d\n", lastres.p, lastres.unit);
+		return 0;
+	}
+
+	log(ERROR, "invalid expression\n");
+	return -1;
+}
 
 static void binprint(maxuint_t n)
 {
-	int count = 127;
-	char binstr[129] = {0};
+	int count = MAX_BITS - 1;
+	char binstr[MAX_BITS + 1] = {0};
 
 	if (!n) {
 		printf("0b0");
@@ -110,8 +260,7 @@ static char *getstr_u128(maxuint_t n, char *buf)
 
 static char *getstr_f128(maxfloat_t val, char *buf)
 {
-	int n = quadmath_snprintf(buf, FLOAT_BUF_LEN,
-				  "%#*.10Qe", FLOAT_WIDTH, val);
+	int n = snprintf(buf, FLOAT_BUF_LEN, "%#*.10Le", FLOAT_WIDTH, val);
 
 	buf[n] = '\0';
 	return buf;
@@ -135,18 +284,6 @@ static void printhex_u128(maxuint_t n)
 		printf("0x%llx", (ull)n);
 }
 
-#if 0
-static char *strtolower(char *buf)
-{
-	char *p = buf;
-
-	for (; *p; ++p)
-		*p = tolower(*p);
-
-	return buf;
-}
-#endif
-
 /* This function adds check for binary input to strtoul() */
 static ulong strtoul_b(char *token)
 {
@@ -155,7 +292,7 @@ static ulong strtoul_b(char *token)
 	/* NOTE: no NULL check here! */
 
 	if (strlen(token) > 2 && token[0] == '0' &&
-			(token[1] == 'b' || token[1] == 'B')) {
+	    (token[1] == 'b' || token[1] == 'B')) {
 		base = 2;
 	}
 
@@ -170,27 +307,138 @@ static ull strtoull_b(char *token)
 	/* NOTE: no NULL check here! */
 
 	if (strlen(token) > 2 && token[0] == '0' &&
-			(token[1] == 'b' || token[1] == 'B')) {
+	    (token[1] == 'b' || token[1] == 'B')) {
 		base = 2;
 	}
 
 	return strtoull(token + base, NULL, base);
 }
 
-#if 0
-/* This function adds check for binary input to strtoull() */
-static maxuint_t strtoval_b(char *token)
+/* Converts a char to unsigned int according to base */
+static bool ischarvalid(char ch, uint base, uint *val)
 {
-	return 0;
+	if (base == 2)
+	{
+		if (ch == '0' || ch == '1') {
+			*val = ch - '0';
+			return TRUE;
+		}
+	} else if (base == 16) {
+		if (ch >= '0' && ch <= '9') {
+			*val = ch - '0';
+			return TRUE;
+		}
+
+		if (ch >= 'a' && ch <= 'f') {
+			*val = (ch - 'a') + 10;
+			return TRUE;
+		}
+
+		if (ch >= 'A' && ch <= 'F') {
+			*val = (ch - 'A') + 10;
+			return TRUE;
+		}
+	} else if (base == 10) {
+		if (ch >= '0' && ch <= '9') {
+			*val = ch - '0';
+			return TRUE;
+		}
+	}
+
+	return FALSE;
+}
+
+/*
+ * Converts a non-floating representing string to maxuint_t
+ */
+static maxuint_t strtouquad(char *token, char **pch)
+{
+	*pch = PASSED;
+
+	if (!token || !*token) {
+		*pch = FAILED;
+		return 0;
+	}
+
+	char *ptr;
+	maxuint_t val = 0, prevval = 0;
+	uint base = 10, multiplier, digit, bits_used = 0;
+	uint max_bit_len = sizeof(maxuint_t) << 3;
+
+	if (token[0] == '0') {
+		if (token[1] == 'b' || token[1] == 'B') { /* binary */
+			base = 2;
+			multiplier = 1;
+		} else if (token[1] == 'x' || token[1] == 'X') { /* hex */
+			base = 16;
+			multiplier = 4;
+		}
+	}
+
+	if (base == 2 || base == 16) {
+		ptr = token + 2;
+
+		if (!*ptr) {
+			*pch = FAILED;
+			return 0;
+		}
+
+		while (*ptr && *ptr == '0')
+			++ptr;
+
+		if (!*ptr)
+			return 0;
+
+		while (*ptr) {
+			if (bits_used == max_bit_len || !ischarvalid(*ptr, base, &digit)) {
+				*pch = FAILED;
+				return 0;
+			}
+
+			val = (val << multiplier) + digit;
+
+			++bits_used;
+			++ptr;
+		}
+
+		return val;
+	}
+
+	/* Try base 10 for any other pattern */
+	ptr = token;
+	while (*ptr && *ptr == '0')
+		++ptr;
+
+	if (!*ptr)
+		return 0;
+
+	while (*ptr) {
+		if (!ischarvalid(*ptr, base, &digit)) {
+			*pch = FAILED;
+			return 0;
+		}
+
+		val = (val * 10) + digit;
+
+		/* Try to to detect overflow */
+		if (val < prevval) {
+			*pch = FAILED;
+			return 0;
+		}
+
+		prevval = val;
+		++ptr;
+	}
+
+	return val;
 }
-#endif
 
 static maxuint_t convertbyte(char *buf, int *ret)
 {
 	maxfloat_t val;
 	char *pch;
-	/* Convert and print in bytes */
-	maxuint_t bytes = (maxuint_t)strtoflt128(buf, &pch);
+	/* Convert and print in bytes (cannot be in float) */
+	maxuint_t bytes = strtouquad(buf, &pch);
 	if (*pch) {
 		*ret = -1;
 		return 0;
@@ -198,7 +446,7 @@ static maxuint_t convertbyte(char *buf, int *ret)
 		*ret = 0;
 
 	if (minimal) {
-		printf("%s\n", getstr_u128(bytes, uint_buf));
+		printf("%s B\n", getstr_u128(bytes, uint_buf));
 		return bytes;
 	}
 
@@ -240,7 +488,7 @@ static maxuint_t convertbyte(char *buf, int *ret)
 static maxuint_t convertkib(char *buf, int *ret)
 {
 	char *pch;
-	maxfloat_t val, kib = strtoflt128(buf, &pch);
+	maxfloat_t val, kib = strtold(buf, &pch);
 	if (*pch) {
 		*ret = -1;
 		return 0;
@@ -250,7 +498,7 @@ static maxuint_t convertkib(char *buf, int *ret)
 	maxuint_t bytes = (maxuint_t)(kib * 1024);
 
 	if (minimal) {
-		printf("%s\n", getstr_u128(bytes, uint_buf));
+		printf("%s B\n", getstr_u128(bytes, uint_buf));
 		return bytes;
 	}
 
@@ -287,7 +535,7 @@ static maxuint_t convertkib(char *buf, int *ret)
 static maxuint_t convertmib(char *buf, int *ret)
 {
 	char *pch;
-	maxfloat_t val, mib = strtoflt128(buf, &pch);
+	maxfloat_t val, mib = strtold(buf, &pch);
 	if (*pch) {
 		*ret = -1;
 		return 0;
@@ -297,7 +545,7 @@ static maxuint_t convertmib(char *buf, int *ret)
 	maxuint_t bytes = (maxuint_t)(mib * (1 << 20));
 
 	if (minimal) {
-		printf("%s\n", getstr_u128(bytes, uint_buf));
+		printf("%s B\n", getstr_u128(bytes, uint_buf));
 		return bytes;
 	}
 
@@ -334,7 +582,7 @@ static maxuint_t convertmib(char *buf, int *ret)
 static maxuint_t convertgib(char *buf, int *ret)
 {
 	char *pch;
-	maxfloat_t val, gib = strtoflt128(buf, &pch);
+	maxfloat_t val, gib = strtold(buf, &pch);
 	if (*pch) {
 		*ret = -1;
 		return 0;
@@ -344,7 +592,7 @@ static maxuint_t convertgib(char *buf, int *ret)
 	maxuint_t bytes = (maxuint_t)(gib * (1 << 30));
 
 	if (minimal) {
-		printf("%s\n", getstr_u128(bytes, uint_buf));
+		printf("%s B\n", getstr_u128(bytes, uint_buf));
 		return bytes;
 	}
 
@@ -381,7 +629,7 @@ static maxuint_t convertgib(char *buf, int *ret)
 static maxuint_t converttib(char *buf, int *ret)
 {
 	char *pch;
-	maxfloat_t val, tib = strtoflt128(buf, &pch);
+	maxfloat_t val, tib = strtold(buf, &pch);
 	if (*pch) {
 		*ret = -1;
 		return 0;
@@ -391,7 +639,7 @@ static maxuint_t converttib(char *buf, int *ret)
 	maxuint_t bytes = (maxuint_t)(tib * ((maxuint_t)1 << 40));
 
 	if (minimal) {
-		printf("%s\n", getstr_u128(bytes, uint_buf));
+		printf("%s B\n", getstr_u128(bytes, uint_buf));
 		return bytes;
 	}
 
@@ -428,7 +676,7 @@ static maxuint_t converttib(char *buf, int *ret)
 static maxuint_t convertkb(char *buf, int *ret)
 {
 	char *pch;
-	maxfloat_t val, kb = strtoflt128(buf, &pch);
+	maxfloat_t val, kb = strtold(buf, &pch);
 	if (*pch) {
 		*ret = -1;
 		return 0;
@@ -438,7 +686,7 @@ static maxuint_t convertkb(char *buf, int *ret)
 	maxuint_t bytes = (maxuint_t)(kb * 1000);
 
 	if (minimal) {
-		printf("%s\n", getstr_u128(bytes, uint_buf));
+		printf("%s B\n", getstr_u128(bytes, uint_buf));
 		return bytes;
 	}
 
@@ -475,7 +723,7 @@ static maxuint_t convertkb(char *buf, int *ret)
 static maxuint_t convertmb(char *buf, int *ret)
 {
 	char *pch;
-	maxfloat_t val, mb = strtoflt128(buf, &pch);
+	maxfloat_t val, mb = strtold(buf, &pch);
 	if (*pch) {
 		*ret = -1;
 		return 0;
@@ -485,7 +733,7 @@ static maxuint_t convertmb(char *buf, int *ret)
 	maxuint_t bytes = (maxuint_t)(mb * 1000000);
 
 	if (minimal) {
-		printf("%s\n", getstr_u128(bytes, uint_buf));
+		printf("%s B\n", getstr_u128(bytes, uint_buf));
 		return bytes;
 	}
 
@@ -522,7 +770,7 @@ static maxuint_t convertmb(char *buf, int *ret)
 static maxuint_t convertgb(char *buf, int *ret)
 {
 	char *pch;
-	maxfloat_t val, gb = strtoflt128(buf, &pch);
+	maxfloat_t val, gb = strtold(buf, &pch);
 	if (*pch) {
 		*ret = -1;
 		return 0;
@@ -532,7 +780,7 @@ static maxuint_t convertgb(char *buf, int *ret)
 	maxuint_t bytes = (maxuint_t)(gb * 1000000000);
 
 	if (minimal) {
-		printf("%s\n", getstr_u128(bytes, uint_buf));
+		printf("%s B\n", getstr_u128(bytes, uint_buf));
 		return bytes;
 	}
 
@@ -569,7 +817,7 @@ static maxuint_t convertgb(char *buf, int *ret)
 static maxuint_t converttb(char *buf, int *ret)
 {
 	char *pch;
-	maxfloat_t val, tb = strtoflt128(buf, &pch);
+	maxfloat_t val, tb = strtold(buf, &pch);
 	if (*pch) {
 		*ret = -1;
 		return 0;
@@ -579,7 +827,7 @@ static maxuint_t converttb(char *buf, int *ret)
 	maxuint_t bytes = (__uint128_t)(tb * 1000000000000);
 
 	if (minimal) {
-		printf("%s\n", getstr_u128(bytes, uint_buf));
+		printf("%s B\n", getstr_u128(bytes, uint_buf));
 		return bytes;
 	}
 
@@ -697,7 +945,7 @@ static bool lba2chs(char *lba, t_chs *p_chs)
 {
 	int token_no = 0;
 	char *ptr, *token;
-	maxuint_t param[3] = {0, MAX_HEAD, MAX_SECTOR};
+	ull param[3] = {0, MAX_HEAD, MAX_SECTOR};
 
 	ptr = token = lba;
 
@@ -766,20 +1014,47 @@ static bool lba2chs(char *lba, t_chs *p_chs)
 	return TRUE;
 }
 
-static void usage(void)
+static void show_basic_sizes()
 {
-	printf("usage: bcal [-c N] [-f FORMAT] [-s bytes] [-d] [-h]\n\
-            [expression] [N unit] \n\n\
-Perform storage conversions and calculations.\n\n\
+	printf("char       : %lu\n"
+		"short      : %lu\n"
+		"int        : %lu\n"
+		"long       : %lu\n"
+		"long long  : %lu\n"
+#ifdef __SIZEOF_INT128__
+		"__int128_t : %lu\n"
+#else
+		"__int64_t  : %lu\n"
+#endif
+		"float      : %lu\n"
+		"double     : %lu\n"
+		"long double: %lu\n",
+		sizeof(unsigned char),
+		sizeof(unsigned short),
+		sizeof(unsigned int),
+		sizeof(unsigned long),
+		sizeof(unsigned long long),
+		sizeof(maxuint_t),
+		sizeof(float),
+		sizeof(double),
+		sizeof(long double));
+}
+
+static void usage()
+{
+	printf("usage: bcal [-c N] [-f FORMAT] [-s bytes] [-m] [-b] [-d] [-h]\n\
+            [expression] [N [unit]] \n\n\
+Storage expression calculator.\n\n\
 positional arguments:\n\
  expression  evaluate storage arithmetic expression\n\
-             +, -, *, / with decimal inputs supported\n\
+             +, -, *, / operators with decimal or hex operands\n\
              unit can be multiplied or divided by +ve integers\n\
-             units can be added or subtracted from each other\n\
+             +, -, / work with two unit operands\n\
              Examples:\n\
                bcal \"(5kb+2mb)/3\"\n\
                bcal \"5 tb / 12\"\n\
                bcal \"2.5mb*3\"\n\
+               bcal \"(2giB * 2) / 2kib\"\n\
  N [unit]    capacity in B/KiB/MiB/GiB/TiB/kB/MB/GB/TB\n\
              see https://wiki.ubuntu.com/UnitsPolicy\n\
              default unit is B (byte), case is ignored\n\
@@ -802,15 +1077,16 @@ optional arguments:\n\
              default MAX_HEAD: 16, default MAX_SECTOR: 63\n\
  -s bytes    sector size [default 512]\n\
  -m          show minimal output (e.g. decimal bytes)\n\
+ -b          list sizes of basic data types in bytes\n\
  -d          enable debug information and logs\n\
  -h          show this help and exit\n\n\
 Version %s\n\
-Copyright © 2016-2017 Arun Prakash Jana <engineerarun@gmail.com>\n\
+Copyright © 2016-2018 Arun Prakash Jana <engineerarun@gmail.com>\n\
 License: GPLv3\n\
 Webpage: https://github.com/jarun/bcal\n", VERSION);
 }
 
-static int xstricmp(const char *s1, const char *s2)
+static int bstricmp(const char *s1, const char *s2)
 {
 	while (*s1 && (tolower(*s1) == tolower(*s2))) {
 		++s1;
@@ -835,15 +1111,24 @@ static maxuint_t unitconv(Data bunit, char *isunit, int *out)
 		log(ERROR, "invalid token\n");
 		*out = -1;
 		return 0;
-	} else
-		*out = 0;
+	}
 
+	*out = 0;
 	len = strlen(numstr) - 1;
 	if (isdigit(numstr[len])) {
+		char *pc = NULL;
+		maxuint_t r;
+
 		/* ensure this is not the result of a previous operation */
 		if (*isunit != 1)
 			*isunit = 0;
-		return strtoull(numstr, NULL, 0);
+		r = strtoull(numstr, &pc, 0);
+		if (*numstr != '\0' && *pc == '\0')
+			return r;
+
+		log(ERROR, "invalid token: %s\n", pc);
+		*out = -1;
+		return 0;
 	}
 
 	while (isalpha(numstr[len]))
@@ -853,11 +1138,15 @@ static maxuint_t unitconv(Data bunit, char *isunit, int *out)
 
 	count = ARRAY_SIZE(units);
 	while (--count >= 0)
-		if (!xstricmp(units[count], punit))
+		if (!bstricmp(units[count], punit))
 			break;
 
 	if (count == -1) {
-		log(ERROR, "unknown unit\n");
+		if (minimal)
+			log(ERROR, "unknown unit\n");
+		else
+			try_bc();
+
 		*out = -1;
 		return 0;
 	}
@@ -868,48 +1157,48 @@ static maxuint_t unitconv(Data bunit, char *isunit, int *out)
 	switch (count) {
 	case 0: {
 		maxuint_t bytes;
-		bytes = (maxuint_t)strtoflt128(numstr, &punit);
+		bytes = (maxuint_t)strtold(numstr, &punit);
 		if (*punit)
 			goto error;
 		return bytes;
 		}
 	case 1: /* Kibibyte */
-		byte_metric = strtoflt128(numstr, &punit);
+		byte_metric = strtold(numstr, &punit);
 		if (*punit)
 			goto error;
 		return (maxuint_t)(byte_metric * 1024);
 	case 2: /* Mebibyte */
-		byte_metric = strtoflt128(numstr, &punit);
+		byte_metric = strtold(numstr, &punit);
 		if (*punit)
 			goto error;
 		return (maxuint_t)(byte_metric * (1 << 20));
 	case 3: /* Gibibyte */
-		byte_metric = strtoflt128(numstr, &punit);
+		byte_metric = strtold(numstr, &punit);
 		if (*punit)
 			goto error;
 		return (maxuint_t)(byte_metric * (1 << 30));
 	case 4: /* Tebibyte */
-		byte_metric = strtoflt128(numstr, &punit);
+		byte_metric = strtold(numstr, &punit);
 		if (*punit)
 			goto error;
 		return (maxuint_t)(byte_metric * ((maxuint_t)1 << 40));
 	case 5: /* Kilobyte */
-		byte_metric = strtoflt128(numstr, &punit);
+		byte_metric = strtold(numstr, &punit);
 		if (*punit)
 			goto error;
 		return (maxuint_t)(byte_metric * 1000);
 	case 6: /* Megabyte */
-		byte_metric = strtoflt128(numstr, &punit);
+		byte_metric = strtold(numstr, &punit);
 		if (*punit)
 			goto error;
 		return (maxuint_t)(byte_metric * 1000000);
 	case 7: /* Gigabyte */
-		byte_metric = strtoflt128(numstr, &punit);
+		byte_metric = strtold(numstr, &punit);
 		if (*punit)
 			goto error;
 		return (maxuint_t)(byte_metric * 1000000000);
 	case 8: /* Terabyte */
-		byte_metric = strtoflt128(numstr, &punit);
+		byte_metric = strtold(numstr, &punit);
 		if (*punit)
 			goto error;
 		return (maxuint_t)(byte_metric * 1000000000000);
@@ -948,7 +1237,7 @@ static int infix2postfix(char *exp, queue **resf, queue **resr)
 
 	while (token) {
 		/* Copy argument to string part of the structure */
-		strcpy(tokenData.p, token);
+		bstrlcpy(tokenData.p, token, NUM_LEN);
 
 		switch (token[0]) {
 		case '+':
@@ -985,6 +1274,16 @@ static int infix2postfix(char *exp, queue **resf, queue **resr)
 			pop(&op, &ct);
 			--balanced;
 			break;
+		case 'r':
+			if (lastres.p[0] == '\0') {
+				log(ERROR, "no result stored\n");
+				emptystack(&op);
+				cleanqueue(resf);
+				return -1;
+			}
+
+			enqueue(resf, resr, lastres);
+			break;
 		default:
 			/* Enqueue operands */
 			enqueue(resf, resr, tokenData);
@@ -1008,7 +1307,34 @@ static int infix2postfix(char *exp, queue **resf, queue **resr)
 	return 0;
 }
 
+/*
+ * Checks for underflow in division
+ * Returns:
+ *  0 - no issues
+ * -1 - underflow
+ */
+static int validate_div(maxuint_t dividend, maxuint_t divisor, maxuint_t quotient)
+{
+	if (divisor * quotient < dividend) {
+		log(WARNING, "result truncated\n");
+
+		if (cur_loglevel == DEBUG) {
+			printhex_u128(dividend);
+			printf(" (dividend)\n");
+			printhex_u128(divisor);
+			printf(" (divisor)\n");
+			printhex_u128(quotient);
+			printf(" (quotient)\n");
+		}
+
+		return -1;
+	}
+
+	return 0;
+}
+
 /* Evaluates Postfix Expression
+ * Numeric result if out parameter holds 1
  * Failure if out parameter holds -1
  */
 static maxuint_t eval(queue **front, queue **rear, int *out)
@@ -1091,11 +1417,23 @@ static maxuint_t eval(queue **front, queue **rear, int *out)
 				log(ERROR, "unit mismatch in *\n");
 				goto error;
 			case '/':
-				/* Divisor must be a non-unit */
-				if (!raw_b.unit) {
+				if (b == 0) {
+					log(ERROR, "division by 0\n");
+					goto error;
+				}
+
+				if (raw_a.unit && raw_b.unit) {
+					c = a / b;
+					raw_c.unit = 0;
+
+					validate_div(a, b, c);
+					break;
+				} else if (!raw_b.unit) {
 					c = a / b;
 					if (raw_a.unit)
 						raw_c.unit = 1;
+
+					validate_div(a, b, c);
 					break;
 				}
 
@@ -1104,7 +1442,7 @@ static maxuint_t eval(queue **front, queue **rear, int *out)
 			}
 
 			/* Convert to string */
-			strcpy(raw_c.p, getstr_u128(c, uint_buf));
+			bstrlcpy(raw_c.p, getstr_u128(c, uint_buf), NUM_LEN);
 			/* Push to stack */
 			push(&est, raw_c);
 
@@ -1122,6 +1460,9 @@ static maxuint_t eval(queue **front, queue **rear, int *out)
 		goto error;
 	}
 
+	if (res.unit == 0)
+		*out = 1;
+
 	/* Convert string to integer */
 	return strtoull(res.p, NULL, 0);
 
@@ -1154,6 +1495,7 @@ static int isoperator(char c)
 	}
 }
 
+#if 0
 /* Check if valid storage arithmetic expression */
 static int checkexp(char *exp)
 {
@@ -1166,8 +1508,9 @@ static int checkexp(char *exp)
 
 	return 0;
 }
+#endif
 
-/* Trim all whitespace from both ends, in place */
+/* Trim ending newline and whitespace from both ends, in place */
 static void strstrip(char *s)
 {
 	if (!s || !*s)
@@ -1175,12 +1518,14 @@ static void strstrip(char *s)
 
 	int len = strlen(s) - 1;
 
-	while (len >= 0 && isspace(s[len]))
+	if (s[len] == '\n')
+		--len;
+	while (len >= 0 && (isspace(s[len]) || s[len] == '\"' || s[len] == '\''))
 		--len;
 	s[len + 1] = '\0';
 
 	len = 0;
-	while (s[len] && isspace(s[len]))
+	while (s[len] && (isspace(s[len]) || s[len] == '\"' || s[len] == '\''))
 		++len;
 
 	if (len) {
@@ -1213,30 +1558,41 @@ static void removeinnerspaces(char *s)
 /* Make the expression compatible with parsing by
  * inserting/removing space between arguments
  */
-static char *fixexpr(char *exp)
+static char *fixexpr(char *exp, int *unitless)
 {
+	*unitless = 0;
+
 	strstrip(exp);
 	removeinnerspaces(exp);
 
+#if 0
 	if (!checkexp(exp)) {
-		unitless = 1;
+		log(DEBUG, "no unit in expression [%s]\n", exp);
+		*unitless = 1;
 		return NULL;
 	}
+#endif
 
 	int i = 0, j = 0;
-	char *parsed = (char *)malloc(2 * strlen(exp) * sizeof(char));
+	char *parsed = (char *)calloc(1, 2 * strlen(exp) * sizeof(char));
 	char prev = '(';
 
 	log(DEBUG, "exp (%s)\n", exp);
 
 	while (exp[i] != '\0') {
+		if (exp[i] == '{' || exp[i] == '}' || exp[i] == '[' || exp[i] == ']') {
+			log(ERROR, "first brackets only\n");
+			free(parsed);
+			return NULL;
+		}
+
 		if (exp[i] == '-' && (issign(prev) || prev == '(')) {
 			log(ERROR, "negative token\n");
 			free(parsed);
 			return NULL;
 		}
 
-		if (isoperator(exp[i]) && isalpha(exp[i + 1])) {
+		if (isoperator(exp[i]) && isalpha(exp[i + 1]) && (exp[i + 1] != 'r')) {
 			log(ERROR, "invalid expression\n");
 			free(parsed);
 			return NULL;
@@ -1245,7 +1601,8 @@ static char *fixexpr(char *exp)
 		if ((isdigit(exp[i]) && isoperator(exp[i + 1])) ||
 		    (isoperator(exp[i]) && (isdigit(exp[i + 1]) ||
 		     isoperator(exp[i + 1]))) ||
-		    (isalpha(exp[i]) && isoperator(exp[i + 1]))) {
+		    (isalpha(exp[i]) && isoperator(exp[i + 1])) ||
+		    (isoperator(exp[i]) && (exp[i + 1] == 'r'))) {
 			parsed[j] = exp[i];
 			++j;
 			parsed[j] = ' ';
@@ -1273,9 +1630,9 @@ static char *fixexpr(char *exp)
 		++i;
 
 	if (!parsed[i]) {
-		log(DEBUG, "noop expression\n");
+		log(DEBUG, "no operator in expression [%s]\n", parsed);
 		free(parsed);
-		unitless = 1;
+		*unitless = 1;
 		return NULL;
 	}
 
@@ -1306,7 +1663,7 @@ static int convertunit(char *value, char *unit, ulong sectorsz)
 
 		if (unitchars) {
 			while (--count >= 0)
-				if (!xstricmp(units[count], value + len))
+				if (!bstricmp(units[count], value + len))
 					break;
 
 			if (count == -1) {
@@ -1321,7 +1678,7 @@ static int convertunit(char *value, char *unit, ulong sectorsz)
 		strstrip(unit);
 
 		while (--count >= 0)
-			if (!xstricmp(units[count], unit))
+			if (!bstricmp(units[count], unit))
 				break;
 
 		if (count == -1) {
@@ -1332,7 +1689,7 @@ static int convertunit(char *value, char *unit, ulong sectorsz)
 
 	log(DEBUG, "%s %s\n", value, units[count]);
 
-	if (!minimal)
+	if (!minimal && unit)
 		printf("\033[1mUNIT CONVERSION\033[0m\n");
 
 	switch (count) {
@@ -1369,10 +1726,18 @@ static int convertunit(char *value, char *unit, ulong sectorsz)
 	}
 
 	if (ret == -1) {
-		log(ERROR, "malformed input\n");
+		if (minimal || unit) /* For running python test cases */
+			log(ERROR, "malformed input\n");
+		else
+			return try_bc();
+
 		return -1;
 	}
 
+	bstrlcpy(lastres.p, getstr_u128(bytes, uint_buf), UINT_BUF_LEN);
+	lastres.unit = 1;
+	log(DEBUG, "result: %s %d\n", lastres.p, lastres.unit);
+
 	if (minimal)
 		return 0;
 
@@ -1396,15 +1761,20 @@ static int convertunit(char *value, char *unit, ulong sectorsz)
 	return 0;
 }
 
-static int evaluate(char *exp)
+static int evaluate(char *exp, ulong sectorsz)
 {
 	int ret = 0;
 	maxuint_t bytes = 0;
 	queue *front = NULL, *rear = NULL;
-	char *expr = fixexpr(exp);	 /* Make parsing compatible */
+	char *expr = fixexpr(exp, &ret);  /* Make parsing compatible */
+	char *ptr;
+
+	if (expr == NULL) {
+		if (ret)
+			return convertunit(exp, NULL, sectorsz);
 
-	if (expr == NULL)
 		return -1;
+	}
 
 	ret = infix2postfix(expr, &front, &rear);
 	free(expr);
@@ -1414,18 +1784,17 @@ static int evaluate(char *exp)
 	bytes = eval(&front, &rear, &ret);  /* Evaluate Expression */
 	if (ret == -1)
 		return -1;
-
-	if (minimal) {
-		convertbyte(getstr_u128(bytes, uint_buf), &ret);
-		if (ret == -1) {
-			log(ERROR, "malformed input\n");
-			return -1;
-		}
-
+	else if (ret == 1) {
+		ptr = getstr_u128(bytes, uint_buf);
+		printf("%s\n", ptr);
+		bstrlcpy(lastres.p, getstr_u128(bytes, uint_buf), UINT_BUF_LEN);
+		lastres.unit = 0;
+		log(DEBUG, "result1: %s %d\n", lastres.p, lastres.unit);
 		return 0;
 	}
 
-	printf("\033[1mRESULT\033[0m\n");
+	if (!(minimal || repl))
+		printf("\033[1mRESULT\033[0m\n");
 
 	convertbyte(getstr_u128(bytes, uint_buf), &ret);
 	if (ret == -1) {
@@ -1433,8 +1802,15 @@ static int evaluate(char *exp)
 		return -1;
 	}
 
-	printf("\nADDRESS\n (d) %s\n (h) ",
-		getstr_u128(bytes, uint_buf));
+	ptr = getstr_u128(bytes, uint_buf);
+	bstrlcpy(lastres.p, ptr, UINT_BUF_LEN);
+	lastres.unit = 1;
+	log(DEBUG, "result2: %s %d\n", lastres.p, lastres.unit);
+
+	if (minimal)
+		return 0;
+
+	printf("\nADDRESS\n (d) %s\n (h) ", ptr);
 	printhex_u128(bytes);
 	printf("\n");
 
@@ -1443,22 +1819,31 @@ static int evaluate(char *exp)
 
 int main(int argc, char **argv)
 {
-	int opt = 0;
+	int opt = 0, operation = 0;
 	ulong sectorsz = SECTOR_SIZE;
 
 	opterr = 0;
+	rl_bind_key('\t', rl_insert);
 
-	while ((opt = getopt(argc, argv, "c:df:hms:")) != -1) {
+	while ((opt = getopt(argc, argv, "bc:df:hms:")) != -1) {
 		switch (opt) {
 		case 'c':
+		{
+			char *pch;
+			operation = 1;
+
 			if (*optarg == '-') {
 				log(ERROR, "N must be >= 0\n");
 				return -1;
 			}
 
-			printf("\033[1mBASE CONVERSION\033[0m\n");
-			maxuint_t val = strtoull_b(optarg);
+			maxuint_t val = strtouquad(optarg, &pch);
+			if (*pch) {
+				log(ERROR, "invalid input\n");
+				return -1;
+			}
 
+			printf("\033[1mBASE CONVERSION\033[0m\n");
 			printf(" (b) ");
 			binprint(val);
 			printf("\n (d) %s\n (h) ",
@@ -1466,7 +1851,10 @@ int main(int argc, char **argv)
 			printhex_u128(val);
 			printf("\n\n");
 			break;
+		}
 		case 'f':
+			operation = 1;
+
 			if (tolower(*optarg) == 'c') {
 				maxuint_t lba = 0;
 
@@ -1498,9 +1886,14 @@ int main(int argc, char **argv)
 			}
 			sectorsz = strtoul_b(optarg);
 			break;
+		case 'b':
+			show_basic_sizes();
+			break;
 		case 'd':
 			cur_loglevel = DEBUG;
 			log(DEBUG, "bcal v%s\n", VERSION);
+			log(DEBUG, "maxuint_t is %lu bytes\n", sizeof(maxuint_t));
+
 			break;
 		case 'h':
 			usage();
@@ -1511,9 +1904,69 @@ int main(int argc, char **argv)
 		}
 	}
 
-	if (argc == 1 && optind == 1) {
-		usage();
-		return -1;
+	log(DEBUG, "argc %d, optind %d\n", argc, optind);
+
+	if (!operation && (argc == optind)) {
+		char *tmp = NULL;
+		repl = 1;
+		int enters = 0;
+
+		printf("q/double Enter -> quit, ? -> help\n");
+		while ((tmp = readline("bcal> ")) != NULL) {
+			if (tmp[0] == '\0') {
+				if (enters == 1)
+					break;
+
+				++enters;
+				continue;
+			}
+
+			enters = 0;
+
+			strstrip(tmp);
+			if (tmp[0] == '\0') {
+				free(tmp);
+				continue;
+			}
+
+			curexpr = tmp;
+			add_history(tmp);
+
+			if (tmp[1] == '\0') {
+				/* Show the last stored result */
+				if (tmp[0] == 'r') {
+					if (lastres.p[0] == '\0')
+						printf("no result stored\n");
+					else {
+						printf("r = %s ", lastres.p);
+						if (lastres.unit)
+							printf("B");
+						printf("\n");
+					}
+
+					free(tmp);
+					continue;
+				}
+
+				if (tmp[0] == '?') {
+					free(tmp);
+					usage();
+					continue;
+				}
+
+				if (tmp[0] == 'q') {
+					free(tmp);
+					break;
+				}
+			}
+
+			/* Evaluate the expression */
+			evaluate(tmp, sectorsz);
+
+			free(tmp);
+		}
+
+		return 0;
 	}
 
 	/* Conversion */
@@ -1522,14 +1975,10 @@ int main(int argc, char **argv)
 			return -1;
 
 	/*Arithmetic operation*/
-	if (argc - optind == 1)
-		if ((evaluate(argv[optind]) == -1) && !unitless)
-			return -1;
+	if (argc - optind == 1) {
+		curexpr = argv[optind];
+		return evaluate(argv[optind], sectorsz);
+	}
 
-	/* Possibly a unitless byte */
-	if (unitless)
-		if (convertunit(argv[optind], NULL, sectorsz) == -1)
-			return -1;
-
-	return 0;
+	return -1;
 }
